from typing import Any, Dict

import numpy as np
from rouge_score import rouge_scorer
from tabulate import tabulate
from nltk.stem.snowball import SnowballStemmer
from types import SimpleNamespace
import re


from .helper import log


class RAGEvaluator:
    def __init__(self):
        self.rus_stem = SnowballStemmer("russian")

        self.rouge_scorer = rouge_scorer.RougeScorer(
            ["rouge1", "rougeL"],
            tokenizer=SimpleNamespace(tokenize=self.tokenize_ru),
            use_stemmer=False
        )

    def tokenize_ru(self, text: str):
        tokens = re.findall(r"\w+", text.lower(), flags=re.UNICODE)
        return [self.rus_stem.stem(t) for t in tokens]

    def evaluate_retrieval(self, retrieved_doc_ids, relevant_doc_id):
        metrics = {}

        # hit rate
        metrics["hit_rate"] = 1.0 if relevant_doc_id in retrieved_doc_ids else 0.0

        # mrr
        for i, doc_id in enumerate(retrieved_doc_ids):
            if doc_id == relevant_doc_id:
                metrics["mrr"] = 1.0 / (i + 1)
                break
        else:
            metrics["mrr"] = 0

        return metrics

    def evaluate_generation(self, generated_answer: str, reference_answer: str):
        rouge_scores = self.rouge_scorer.score(generated_answer, reference_answer)
        return {
            "rouge1": rouge_scores["rouge1"].fmeasure,
            "rougeL": rouge_scores["rougeL"].fmeasure,
            "em": self.evaluate_em(generated_answer, reference_answer),
            "substring_match": self.evaluate_substring_match(generated_answer, reference_answer)
        }

    def evaluate_em(self, generated_answer: str, reference_answer: str):
        return 1.0 if generated_answer == reference_answer else 0.0

    def evaluate_substring_match(self, generated_answer: str, reference_answer: str):
        return 1.0 if generated_answer.lower() in reference_answer.lower() else 0.0


class RAGEvaluationResults:
    def __init__(self, individual_results: Dict[str, Any], average_metrics: Dict[str, Any]):
        self.individual_results = individual_results
        self.average_metrics = average_metrics

    @classmethod
    def from_dict(cls, results_dict: Dict[str, Any]):
        return cls(
            individual_results=results_dict['individual_results'],
            average_metrics=results_dict['average_metrics']
        )

    def to_dict(self):
        return {
            "individual_results": self.individual_results,
            "average_metrics": self.average_metrics
        }

    def to_table(self):
        retrieval_metrics = self.average_metrics['retrieval']
        retrieval_table = tabulate([
            ["Hit Rate", retrieval_metrics['hit_rate']],
            ["MRR", retrieval_metrics['mrr']]
        ], headers=["Metric", "Value"], 
        tablefmt="grid",
        floatfmt=".3f")

        generation_metrics = self.average_metrics['generation']
        generation_table = tabulate([
            ["ROUGE-1", generation_metrics['rouge1']],
            ["ROUGE-L", generation_metrics['rougeL']]
        ], headers=["Metric", "Value"], 
        tablefmt="grid",
        floatfmt=".3f")

        log("Retrieval Metrics:")
        log(retrieval_table)
        log("\nGeneration Metrics:")
        log(generation_table)
        

def evaluate_rag_results(results, dataset):
    evaluation_results = {}
    evaluator = RAGEvaluator()

    for i, result in results.items():
        reference_answer = dataset["train"][int(i)]["answer"]

        retrieval_metrics = evaluator.evaluate_retrieval(
            retrieved_doc_ids=result["found_ids"], 
            relevant_doc_id=int(i)
        )

        generation_metrics = evaluator.evaluate_generation(
            generated_answer=result["model_answer"], 
            reference_answer=reference_answer
        )

        evaluation_results[i] = {
            "retrieval": retrieval_metrics,
            "generation": generation_metrics,
        }

    avg_metrics = {"retrieval": {}, "generation": {}}

    for metric in ["hit_rate", "mrr"]:
        avg_metrics["retrieval"][metric] = np.mean(
            [res["retrieval"][metric] for res in evaluation_results.values()]
        )

    for metric in ["rouge1", "rougeL", "em", "substring_match"]:
        avg_metrics["generation"][metric] = np.mean(
            [res["generation"][metric] for res in evaluation_results.values()]
        )

    return RAGEvaluationResults(evaluation_results, avg_metrics)


# RAG pipeline reference implementation

import time
from uuid import uuid4

from .helper import log
from langchain.chains import create_retrieval_chain
from langchain.chains.combine_documents import create_stuff_documents_chain
from langchain_chroma import Chroma
from langchain_core.documents import Document
from langchain_core.prompts import ChatPromptTemplate
from langchain_text_splitters import RecursiveCharacterTextSplitter
from tqdm import tqdm


def init_retriever(
    dataset, embedding_model, top_k=5, chunk_size=500, chunk_overlap=100
):
    log("Initializing retriever")
    vector_store = Chroma(embedding_function=embedding_model)
    log("> Creating vector store")
    start_time = time.time()
    documents = [
        Document(id=i, page_content=f"{item['text']}", metadata={"id": item["id"]})
        for i, item in enumerate(dataset["train"])
    ]
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size, chunk_overlap=chunk_overlap
    )
    splits = text_splitter.split_documents(documents)
    uuids = [str(uuid4()) for _ in range(len(splits))]

    vector_store.reset_collection()
    vector_store.add_documents(documents=splits, ids=uuids)

    retriever = vector_store.as_retriever(search_type="mmr", search_kwargs={"k": top_k})

    log(f"> Vector store created for {len(documents)} documents")
    log(f"> Done for {time.time() - start_time:.2f} seconds")

    return retriever


def init_generation(retriever, model, llm_prompt=""):
    log("Initializing generation chain")
    if not llm_prompt:
        llm_prompt = """Answer a question using the provided context. Return answer only.

<context>
{context[:30000]}
</context>

Question: {input}"""
    prompt = ChatPromptTemplate.from_template(llm_prompt)
    document_chain = create_stuff_documents_chain(model, prompt)
    generation_chain = create_retrieval_chain(retriever, document_chain)
    log("> Done")
    return generation_chain


def prepare_results(data):
    """Prepare your results for evaluation."""
    res = {}
    for item_id, response in data.items():
        res[item_id] = {
            "found_ids": [x.metadata["id"] for x in response["context"]],
            "model_answer": response["answer"],
        }
    return res


def get_results(generation_chain, dataset, skip=None, take=None, write_logs=False, sleep_time=0):
    log("Calculating and preparing results")
    results = {}
    dataset = dataset["train"]
    if skip:
        dataset = dataset.select(range(skip, len(dataset)))
    if take:
        dataset = dataset.select(range(take))
    start_time = time.time()
    for item in tqdm(dataset):
        response = generation_chain.invoke({"input": item["question"]})

        if write_logs:
            print("*" * 80)
            print("id:", item["id"])
            print("Question:", item["question"])
            # print("Answer:", item["answer"])
            print("-----")
            print("Response:", response["answer"])

        results[item["id"]] = response
        if sleep_time > 0:
            time.sleep(sleep_time)
    log(f"Done for {time.time() - start_time:.2f} seconds")
    results = prepare_results(results)
    return results



import json
import os
import uuid
from pathlib import Path

import requests


def save(data, path):
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    json.dump(data, open(path, "w", encoding="utf8"), ensure_ascii=False, indent=4)


def load(path):
    return json.load(open(path, "r", encoding="utf8"))


def submit(path, config, address="localhost:80", is_auto=False, submit_id="", version="", token=""):
    with open(path, "r", encoding="utf-8") as file:
        content = json.load(file)
    # TODO do some content validation before sending

    config["user_guid"] = uuid.uuid4().hex
    config["is_auto"] = is_auto

    with open(path, "rb") as file:
        data = {
            "config": json.dumps(config),
            "submit_id": submit_id,
            "version": version,
            "token": token,
            "filename": os.path.basename(path),
        }
        files = {path: file}
        response = requests.post(
            f"http://{address}/submit/create",
            data=data,
            files=files,
        )

    try:
        res = json.loads(response.content.decode("utf-8"))
    except Exception as e:
        print("Exception occured:", str(e))
        return

    print(f"Done. Your submission id is {res['id']}")



from datetime import datetime
from huggingface_hub import list_repo_refs

from .constants import LOG_FILE

def log(msg):
    print(msg)
    with open(LOG_FILE, "a", encoding="utf-8") as fout:
        date = datetime.now().strftime("%d.%m.%Y %H:%M")
        fout.write(f"[{date}] {msg}\n")

def sort_versions(versions):
    def version_key(v):
        return [int(x) for x in v.split('.')]
    
    return sorted(versions, key=version_key)

def get_ds_versions(repo_id):
    repo_refs = list_repo_refs(repo_id, repo_type="dataset")
    return [ref.name for ref in repo_refs.tags]

def get_latest_version(versions):
    return sort_versions(versions)[-1]




from datasets import load_dataset
from .helper import log, get_latest_version, get_ds_versions

PUBLIC_TEXTS_REPO_ID = "ai-forever/rag-bench-public-texts"
PUBLIC_QUESTIONS_REPO_ID = "ai-forever/rag-bench-public-questions"

def get_texts():
    log("Loading texts dataset")
    return load_dataset(PUBLIC_TEXTS_REPO_ID)

def get_questions():
    log("Loading questions dataset")
    return load_dataset(PUBLIC_QUESTIONS_REPO_ID)

def get_datasets():
    text_versions = get_ds_versions(PUBLIC_TEXTS_REPO_ID)
    question_versions = get_ds_versions(PUBLIC_QUESTIONS_REPO_ID)

    latest_text_version = get_latest_version(text_versions)
    latest_question_version = get_latest_version(question_versions)
    
    log(f"Latest texts version: {latest_text_version}")
    log(f"Latest questions version: {latest_question_version}")

    assert latest_text_version == latest_question_version, "Texts and questions datasets have different versions"

    text_ds = load_dataset(PUBLIC_TEXTS_REPO_ID, revision=latest_text_version)
    question_ds = load_dataset(PUBLIC_QUESTIONS_REPO_ID, revision=latest_question_version)

    log(f"Loaded texts dataset with {len(text_ds['train'])} texts")
    log(f"Loaded questions dataset with {len(question_ds['train'])} questions")

    return text_ds, question_ds, latest_text_version
